import unittest

from ll import url

class URLTestCase(unittest.TestCase):

	def test_parse(self):
		base = "http://a/b/c/d;p?q#f"
		u = url.URL(base)
		self.assertEquals(u.scheme, "http")
		self.assertEquals(u.userinfo, None)
		self.assertEquals(u.host, "a")
		self.assertEquals(u.port, None)
		self.assertEquals(u.hostport, "a")
		self.assertEquals(u.server, "a")
		self.assertEquals(u.authority, "a")
		self.assertEquals(u.reg_name, None)
		self.assertEquals(u.path, "/b/c/d;p")
		self.assertEquals(u.path_segments, [("b",), ("c",), ("d", "p")])
		self.assertEquals(u.isabspath, 1)
		self.assertEquals(u.query, "q")
		self.assertEquals(u.fragment, "f")
		self.assertEquals(u.opaque_part, None)
		self.assertEquals(u.url, base)

	def test_rfc2396join(self):
		base = "http://a/b/c/d;p?q"
		tests = [
			# RFC2396 Section C.1: Normal Examples
			("g:h",           "g:h"),
			("g",             "http://a/b/c/g"),
			("./g",           "http://a/b/c/g"),
			("g/",            "http://a/b/c/g/"),
			("/g",            "http://a/g"),
			("//g",           "http://g"),
			("?y",            "http://a/b/c/?y"),
			("g?y",           "http://a/b/c/g?y"),
			("#s",            "http://a/b/c/d;p?q#s"),
			("g#s",           "http://a/b/c/g#s"),
			("g?y#s",         "http://a/b/c/g?y#s"),
			(";x",            "http://a/b/c/;x"),
			("g;x",           "http://a/b/c/g;x"),
			("g;x?y#s",       "http://a/b/c/g;x?y#s"),
			(".",             "http://a/b/c/"),
			("./",            "http://a/b/c/"),
			("..",            "http://a/b/"),
			("../",           "http://a/b/"),
			("../g",          "http://a/b/g"),
			("../..",         "http://a/"),
			("../../",        "http://a/"),
			("../../g",       "http://a/g"),
			# RFC2396 Section C.2: Abnormal Examples
			("",              "http://a/b/c/d;p?q"),
			("../../../g",    "http://a/../g"),
			("../../../../g", "http://a/../../g"),
			("/./g",          "http://a/./g"),
			("/../g",         "http://a/../g"),
			("g.",            "http://a/b/c/g."),
			(".g",            "http://a/b/c/.g"),
			("g..",           "http://a/b/c/g.."),
			("..g",           "http://a/b/c/..g"),
			("./../g",        "http://a/b/g"),
			("./g/.",         "http://a/b/c/g/"),
			("g/./h",         "http://a/b/c/g/h"),
			("g/../h",        "http://a/b/c/h"),
			("g;x=1/./y",     "http://a/b/c/g;x=1/y"),
			("g;x=1/../y",    "http://a/b/c/y"),
			("g?y/./x",       "http://a/b/c/g?y/./x"),
			("g?y/../x",      "http://a/b/c/g?y/../x"),
			("g#s/./x",       "http://a/b/c/g#s/./x"),
			("g#s/../x",      "http://a/b/c/g#s/../x"),
			("http:g",        "http:g"), # use the validating version here
		]
		baseurl = url.URL(base)
		for (rel, res) in tests:
			relurl = url.URL(rel)
			resurl = url.URL(res)
			self.assertEquals(baseurl/relurl, resurl, "%r/%r is %r, but should be %r" % (baseurl, relurl, baseurl/relurl, resurl))

	def test_join(self):
		tests = [
			("http://test.com/index.html", "impress.html", "http://test.com/impress.html"),
			("http://test.com/index.html", "", "http://test.com/index.html"),
			("/bb/cc/", "http:", "http:"),
			("mailto:x@y.z", "index.html", "index.html"),
			("mailto:x@y.z", "", "mailto:x@y.z"),
			("javascript:return ':/:/:';", "index.html", "index.html"),
			("javascript:document.write('http://foo@bar.com:81/foo;bar/bar;foo?x=y#frag');", "index.html", "index.html"),
			("mailto:x@y", "", "mailto:x@y"),
			("http://test.com/gurk/hurz.gif", "/index.html", "http://test.com/index.html"),
			("http://test.com/gurk/hurz.gif", "../", "http://test.com/"),
			("http://test.com/gurk/hurz.gif", "../gurk.gif?foo=bar#nix", "http://test.com/gurk.gif?foo=bar#nix"),
			("http://test.com/gurk/hurz.gif", "../../gurk.gif?foo=bar#nix", "http://test.com/../gurk.gif?foo=bar#nix"),
			("http://test.com/gurk/hurz.gif", "root:gurk.gif", "root:gurk.gif"),
			("root:gurk.gif", "http://test.com/gurk/hurz.gif", "http://test.com/gurk/hurz.gif"),
			("root:gurk/hurz/hinz.gif", "hinz/kunz.gif", "root:gurk/hurz/hinz/kunz.gif"),
			("root:gurk/hurz/hinz.gif", "root:hinz/kunz.gif", "root:hinz/kunz.gif")
		]

		for (base, rel, res) in tests:
			baseurl = url.URL(base)
			relurl = url.URL(rel)
			resurl = url.URL(res)
			self.assertEquals(baseurl/relurl, resurl, "%r/%r is %r, but should be %r" % (baseurl, relurl, baseurl/relurl, resurl))

	def test_normalize(self):
		tests = [
			("", ""),
			("./", ""),
			("/./", "/"),
			("xx", "xx"),
			("xx/yy", "xx/yy"),
			("xx/..", ""),
			("xx/../.", ""),
			("./xx/..", ""),
			("./xx/../.", ""),
			("xx/./..", ""),
			("xx/yy/..", "xx/"),
			("xx//yy/../..", ""),
			("xx//yy/./..", "xx/"),
			("xx//yy//../", "xx/"),
			("xx/../..//", "../"),
			("xx/.././..", ".."), # ".." parts above the root loose their "directoryness", otherwise this would be "../"
			("xx/.", "xx/"),
			("./xx", "xx"),
			("/xx", "/xx"),
			("/./xx", "/xx"),
		]
		for (u, normu) in tests:
			u = url.URL(u)
			testu = url.URL(u)
			testu.path_segments = url._normalizePath(u.path_segments)
			normu = url.URL(normu)
			self.assertEquals(testu, normu, "%r normalized is %r, but should be %r" % (u, testu, normu))

	def test_repr(self):
		s = "ftp://ftp.livinglogic.de/pub/livinglogic/xist/XIST-42.105.tar.bz2"
		u = url.URL(s)
		self.assertEqual(unicode(u), s)

	def test_relative(self):
		tests = [
			("./", "./", "./"),
			("cc.html", "./", "./"),
			("./cc.html", "./", "./"),
			("file:./cc.html", "file:./", "./"),
			("root:./cc.html", "file:./", "file:./"),
			("root:xist/Documentation.html", "http://server/", "http://server/"),
			("root:cc.html", "root:", "./"),
			("root:cc.html", "./", "./"),
			("cc.html", "#mark", "#mark"),
			("root:cc.html", "root:#mark", "./#mark"),
			("root:cc.html", "#mark", "#mark"),
			("root:cc.html", "root:cc.html#mark", "#mark"),
			("root:cc.html", "root:dd.html#mark", "dd.html#mark"),
			("root:aa/bb/cc.html", "root:", "../../"),
			("", "", ""),
			("http://server/aa/bb.html", "http://server/aa/cc.html", "cc.html"),
			("/aa/bb.html", "/xx.html", "/xx.html"), # we don't handle URLs without scheme
		]
		for (base, rel, res) in tests:
			baseurl = url.URL(base)
			relurl = url.URL(rel)
			resurl = url.URL(res)
			self.assertEquals(relurl.relative(baseurl), resurl, "%r.relative(%r) is %r, but should be %r" % (relurl, baseurl, relurl.relative(baseurl), resurl))

if __name__ == "__main__":
	unittest.main()
