#!/usr/bin/env python

"""
<doc:par>This module contains an <a href="http://www.ietf.org/rfc/rfc2396.txt">RFC2396</a>
compliant implementation of &url;s (see the class <pyref class="URL"><class>URL</class></pyref>)
and file like classes for reading data from &url;s and writing data to &url;s 
(see the classes <pyref class="Resource"><class>Resource</class></pyref>,
<pyref class="ReadResource"><class>ReadResource</class></pyref> and
<pyref class="WriteResource"><class>WriteResource</class></pyref>).</doc:par>
"""

from __future__ import generators

import urllib, urllib2, os, os.path, mimetypes, time, email.Utils, mimetools, cStringIO as StringIO, warnings
from mx import DateTime

from _url import escape as _escape, unescape as _unescape

alpha = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
alphanum = alpha + "0123456789"
mark = "-_.!~*'()"
additionalsafe = "[]"
safe = alphanum + mark + additionalsafe
pathsafe = safe + ":@&=+$," + "|" # add "|" for Windows paths

schemecharfirst = alpha
schemechar = alphanum + "+-."

def _normalizePath(path_segments):
	"""
	<doc:par>Internal helper function for normalizing a path list.</doc:par>
	"""
	# the following should be equivalent to RFC2396, Section 5.2 (6) (c)-(f)
	# with the exception of removing empty path_segments
	new_path_segments = []
	l = len(path_segments)
	for i in xrange(l):
		segment = path_segments[i]
		if segment==(".",) or segment==("",):
			if i==l-1:
				new_path_segments.append(("",))
		elif segment==("..",) and len(new_path_segments) and new_path_segments[-1]!=("..",):
			new_path_segments.pop()
			if i==l-1:
				new_path_segments.append(("",))
		else:
			new_path_segments.append(segment)
	return new_path_segments

class SchemeDefinition(object):
	"""
	<doc:par>A <class>SchemeDefinition</class> instance defines the properties
	of a particular &url; scheme.</doc:par>
	"""
	def __init__(self, scheme, usehierarchy, useserver, usefragment, isLocal=0, isRemote=0, defaultport=None):
		"""
		<doc:par>Create a new <class>SchemeDefinition</class> instance. Arguments are:</doc:par>
		<doc:ulist>
		<doc:item><arg>scheme</arg>: The name of the scheme;</doc:item>
		<doc:item><arg>usehierarchy</arg>: Specifies whether this scheme uses hierarchical &url;s
		or opaque &url;s (i.e. whether <lit>hier_part</lit> or <lit>opaque_part</lit> from the
		&bnf; in <a href="http://www.ietf.org/rfc/rfc2396.txt">RFC2396</a> is used);</doc:item>
		<doc:item><arg>useserver</arg>: Specifies whether this scheme uses an Internet-based server
		<pyref property="authority"><property>authority</property></pyref> component or a registry
		of naming authorities (only for hierarchical &url;s);</doc:item>
		<doc:item><arg>usefragment</arg>: Specifies whether this scheme uses fragments
		(according to the &bnf; in <a href="http://www.ietf.org/rfc/rfc2396.txt">RFC2396</a> 
		every scheme does, but it doesn't make sense for e.g. <lit>"javascript"</lit>,
		<lit>"mailto"</lit> or <lit>"tel"</lit>);</doc:item>
		<doc:item><arg>isLocal</arg>: Specifies whether &url;s with this scheme refer to
		local files;</doc:item>
		<doc:item><arg>isRemote</arg>: Specifies whether &url;s with this scheme refer to
		remote files (there may be schemes which are neither local nor remote, 
		e.g. <lit>"mailto"</lit>);</doc:item>
		<doc:item><arg>defaultport</arg>: The default port for this scheme (only for schemes
		using server based authority).</doc:item>
		</doc:ulist>
		"""
		self.scheme = scheme
		self.usehierarchy = usehierarchy
		self.useserver = useserver
		self.usefragment = usefragment
		self.isLocal = isLocal
		self.isRemote = isRemote
		self.defaultport = defaultport

	def __repr__(self):
		return "<%s instance scheme=%r usehierarchy=%r useserver=%r usefragment=%r at 0x%x>" % (self.__class__.__name__, self.scheme, self.usehierarchy, self.useserver, self.usefragment, id(self))

schemereg = {
	"http": SchemeDefinition("http", usehierarchy=1, useserver=1, usefragment=1, isRemote=1, defaultport=80),
	"https": SchemeDefinition("https", usehierarchy=1, useserver=1, usefragment=1, isRemote=1, defaultport=443),
	"ftp": SchemeDefinition("ftp", usehierarchy=1, useserver=1, usefragment=1, isRemote=1, defaultport=21),
	"file": SchemeDefinition("file", usehierarchy=1, useserver=0, usefragment=1, isLocal=1),
	"root": SchemeDefinition("root", usehierarchy=1, useserver=0, usefragment=1, isLocal=1),
	"javascript": SchemeDefinition("javascript", usehierarchy=0, useserver=0, usefragment=0),
	"mailto": SchemeDefinition("mailto", usehierarchy=0, useserver=0, usefragment=0),
	"tel": SchemeDefinition("tel", usehierarchy=0, useserver=0, usefragment=0),
	"fax": SchemeDefinition("fax", usehierarchy=0, useserver=0, usefragment=0),
}
defaultreg = SchemeDefinition("", usehierarchy=1, useserver=1, isLocal=1, usefragment=1)

def here(scheme="file"):
	"""
	<doc:par>Return the current directory as an <pyref class="URL"><class>URL</class></pyref>.</doc:par>
	"""
	return Dirname(os.getcwd(), scheme)

def home(user="", scheme="file"):
	"""
	<doc:par>Return the home directory of the current user (or the user named <arg>user</arg>, if <arg>user</arg> is specified)
	as an <pyref class="URL"><class>URL</class></pyref>.</doc:par>
	<doc:programlisting>
	&gt;&gt;&gt; url.home()
	URL('file:/home/walter/')
	&gt;&gt;&gt; url.home("andreas")
	URL('file:/home/andreas/')
	</doc:programlisting>
	"""
	return Dirname("~%s" % user, scheme)

def root():
	"""
	<doc:par>Return a blank <lit>root</lit> <pyref class="URL"><class>URL</class></pyref>,
	i.e. <lit>URL("root:")</lit>.</doc:par>
	"""
	return URL("root:")

def Filename(filename, scheme="file"):
	"""
	<doc:par>Turn a filename into an <pyref class="URL"><class>URL</class></pyref>.</doc:par>
	"""
	filename = urllib.pathname2url(os.path.expanduser(filename).encode("utf-8"))
	url = URL(filename)
	url.scheme = scheme
	return url

def Dirname(dirname, scheme="file"):
	"""
	<doc:par>Turns a directory name into an <pyref class="URL"><class>URL</class></pyref>, just like
	<pyref function="Filename"><function>Filename</function></pyref>, but ensures that the path
	is terminated with a <lit>/</lit>.</doc:par>
	"""
	dirname = urllib.pathname2url(os.path.expanduser(dirname).encode("utf-8"))
	if dirname[-1:] != "/":
		dirname += "/"
	url = URL(dirname)
	url.scheme = scheme
	return url

class URL(object):
	"""
	<doc:par></doc:par>
	"""
	def __init__(self, url=None):
		"""
		<doc:par>Create a new <class>URL</class> instance. <arg>url</arg> may be a <class>str</class>
		or <class>unicode</class> instance, or an <class>URL</class> (in which case you'll get of copy
		of <arg>url</arg>), or <lit>None</lit> (which will create an <class>URL</class> referring
		to the <z>current document</z>).</doc:par>
		"""
		self.url = url

	def __clear(self):
		"""
		<doc:par>internal helper method that makes the &url; empty.</doc:par>
		"""
		self.reg = defaultreg
		self._scheme = None
		self._userinfo = None
		self._host = None
		self._port = None
		self._isabspath = 0
		self._path_segments = []
		self._reg_name = None
		self._query = None
		self._opaque_part = None
		self._fragment = None

	def clone(self):
		"""
		<doc:par>Return an identical copy <self/>.</doc:par>
		"""
		return URL(self)

	def _checkscheme(scheme):
		"""
		<doc:par>Check whether <arg>scheme</arg> consists of legal characters.</doc:par>
		"""
		# check if the scheme only has allowed characters
		if scheme[0] not in schemecharfirst:
			return 0
		for c in scheme[1:]:
			if c not in schemechar:
				return 0
		return 1
	_checkscheme = staticmethod(_checkscheme)

	def __getscheme(self):
		return self._scheme
	def __setscheme(self, scheme):
		"""
		<doc:par>The scheme will be converted to lowercase on setting (if <arg>scheme</arg> is not <lit>None</lit>,
		otherwise the scheme will be deleted).</doc:par>
		"""
		if scheme is None:
			self._scheme = None
		else:
			scheme = scheme.lower()
			# check if the scheme only has allowed characters
			if not self._checkscheme(scheme):
				raise ValueError("Illegal scheme char in scheme %r" % (scheme, ))
			self._scheme = scheme
		self.reg = schemereg.get(scheme, defaultreg)
	def __delscheme(self):
		"""
		<doc:par>Deletes the scheme, i.e. makes the &url; relative.</doc:par>
		"""
		self._scheme = None
		self.reg = defaultreg
	scheme = property(
		__getscheme,
		__setscheme,
		__delscheme,
		"""
		<doc:par>the &url; scheme (e.g. <lit>ftp</lit>, <lit>http</lit> or <lit>mailto</lit>).
		The scheme will be <lit>None</lit> if the &url; is a relative one.</doc:par>
		"""
	)

	def __getuserinfo(self):
		return self._userinfo
	def __setuserinfo(self, userinfo):
		self._userinfo = userinfo
	def __deluserinfo(self):
		self._userinfo = None
	userinfo = property(
		__getuserinfo,
		__setuserinfo,
		__deluserinfo,
		"""
		<doc:par>the user info part of the <class>URL</class>; i.e. the <lit>user</lit>
		part of <lit>http://user@www.foo.com:8080/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __gethost(self):
		return self._host
	def __sethost(self, host):
		if host is not None:
			host = host.lower()
		self._host = host
	def __delhost(self):
		self._host = None
	host = property(
		__gethost,
		__sethost,
		__delhost, 
		"""
		<doc:par>the host part of the <class>URL</class>; i.e. the <lit>www.foo.com</lit>
		part of <lit>http://user@www.foo.com:8080/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getport(self):
		return self._port
	def __setport(self, port):
		if port is not None:
			port = int(port)
		self._port = port
	def __delport(self):
		self._port = None
	port = property(
		__getport,
		__setport,
		__delport,
		"""
		<doc:par>the port number of the <class>URL</class> (as an <class>int</class>)
		or <lit>None</lit> if the <class>URL</class> has none. The <lit>8080</lit>
		in <lit>http://user@www.foo.com:8080/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __gethostport(self):
		if self.host is not None:
			hostport = _escape(self.host, safe)
			if self.port is not None:
				hostport += ":%d" % self.port
			return hostport
		else:
			return None
	def __sethostport(self, hostport):
		# find the port number (RFC2396, Section 3.2.2)
		if hostport is None:
			del self.hostport
		else:
			del self.port
			pos = hostport.rfind(":")
			if pos != -1:
				if pos != len(hostport)-1:
					self.port = hostport[pos+1:]
				hostport = hostport[:pos]
			self.host = _unescape(hostport)
	def __delhostport(self):
		del self.host
		del self.port
	hostport = property(
		__gethostport,
		__sethostport,
		__delhostport,
		"""
		<doc:par>the host and (if specified) the port number of the <class>URL</class>,
		i.e. the <lit>www.foo.com:8080</lit> in
		<lit>http://user@www.foo.com:8080/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getserver(self):
		if self.hostport is not None:
			userinfo = self.userinfo
			if userinfo is not None:
				return _escape(userinfo, safe) + "@" + self.hostport
			else:
				return self.hostport
		else:
			return None
	def __setserver(self, server):
		"""
		<doc:par>Setting the server always works even if the current
		<pyref property="scheme"><property>scheme</property></pyref>
		does use <pyref property="opaque_part"><property>opaque_part</property></pyref>
		or <pyref property="reg_name"><property>reg_name</property></pyref>,
		but will be ignored for <pyref property="url"><property>url</property></pyref>.</doc:par>
		"""
		if server is None:
			del self.server
		else:
			# find the userinfo (RFC2396, Section 3.2.2)
			pos = server.find("@")
			if pos != -1:
				self.userinfo = _unescape(server[:pos])
				server = server[pos+1:]
			else:
				del self.userinfo
			self.hostport = server
	def __delserver(self):
		del self.userinfo
		del self.hostport
	server = property(
		__getserver,
		__setserver,
		__delserver,
		"""
		<doc:par>the server part of the <class>URL</class>; i.e. the <lit>user@www.foo.com</lit>
		part of <lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getreg_name(self):
		return self._reg_name
	def __setreg_name(self, reg_name):
		if reg_name is None:
			del self.reg_name
		else:
			self._reg_name = reg_name
	def __delreg_name(self):
		self._reg_name = None
	reg_name = property(
		__getreg_name,
		__setreg_name,
		__delreg_name,
		"""
		<doc:par>The reg_name part of the <class>URL</class> for hierarchical schemes that
		use a name based <pyref property="authority"><property>authority</property></pyref>
		instead of <pyref property="server"><property>server</property></pyref>.</doc:par>
		"""
	)

	def __getauthority(self):
		if self.reg.useserver:
			return self.server
		else:
			return self.reg_name
	def __setauthority(self, authority):
		if self.reg.useserver:
			self.server = authority
		else:
			self.reg_name = authority
	def __delauthority(self):
		if self.reg.useserver:
			del self.server
		else:
			del self.reg_name
	authority = property(
		__getauthority,
		__setauthority,
		__delauthority,
		"""
		<doc:par>The authority part of the <class>URL</class> for hierarchical schemes. Depending
		on the scheme, this is either <pyref property="server"><property>server</property></pyref>
		or <pyref property="reg_name"><property>reg_name</property></pyref>.</doc:par>
		"""
	)

	def __getisabspath(self):
		return (self.authority is not None) or self._isabspath
	def __setisabspath(self, isabspath):
		self._isabspath = isabspath
	isabspath = property(
		__getisabspath,
		__setisabspath,
		None,
		"""
		<doc:par>Specifies whether the path of a hierarchical <class>URL</class> is absolute,
		(i.e. it has a leading <lit>"/"</lit>). Note that the path will always be absolute if an
		<pyref property="authority"><property>authority</property></pyref> is specified.</doc:par>
		"""
	)

	def __getpath_segments(self):
		return self._path_segments
	def __setpath_segments(self, path_segments):
		if path_segments is None:
			del self.path_segments
		else:
			self._path_segments = path_segments
	def __delpath_segments(self):
		"""
		<doc:par>Sets the <property>path_segments</property> to <lit>[("",)]</lit></doc:par>
		"""
		self._path_segments = [("",)]
	path_segments = property(
		__getpath_segments,
		__setpath_segments,
		__delpath_segments,
		"""
		<doc:par>The path segments of a hierarchical <class>URL</class>.
		This is a list of <lit>(<rep>name</rep>, <rep>param</rep>)</lit> tuples).</doc:par>
		<doc:par>If a path segment doesn't have a parameter, the segment tuple will contain
		only the <rep>name</rep> entry. To specify a directory use a segment <lit>("",)</lit>
		as the last list entry. <property>path_segments</property> must have at least one entry.</doc:par>
		"""
	)

	def __getpath(self):
		path = "/".join([";".join([_escape(value, pathsafe) for value in segment ]) for segment in self.path_segments])
		if self.isabspath:
			path = "/" + path
		return path
	def __setpath(self, path):
		if path is None:
			del self.path
		else:
			if path.startswith("/"):
				self.isabspath = 1
				path = path[1:]
			else:
				self.isabspath = 0
			self._path_segments = [ tuple([ _unescape(name) for name in segment.split(";", 1) ]) for segment in path.split("/") ]
	def __delpath(self):
		del self.path_segments
	path = property(
		__getpath,
		__setpath,
		__delpath,
		"""
		<doc:par>The path as a string. This combines
		<pyref property="path_segments"><property>path_segments</property></pyref> and
		<pyref property="isabspath"><property>isabspath</property></pyref>. The <lit>/bar/baz.html;xyzzy</lit>
		part of <lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getfile(self):
		if len(self.path_segments):
			return self.path_segments[-1][0]
		else:
			return None
	def __setfile(self, file):
		"""
		<doc:par>Setting the filename preserves the parameter in the last segment.</doc:par>
		"""
		if file is None:
			del self.file
		if self.path_segments:
			if len(self.path_segments[-1]) == 1:
				self.path_segments[-1] = (file, )
			else:
				self.path_segments[-1] = (file, self.path_segments[-1][1])
		else:
				self.path_segments = [(file,)]
	def __delfile(self):
		"""
		<doc:par>Deleting the filename preserves the parameter in the last segment.</doc:par>
		"""
		if len(self.path_segments):
			if len(self.path_segments[-1]) == 1:
				self.path_segments[-1] = ("", )
			else:
				self.path_segments[-1] = ("", self.path_segments[-1][1])
	file = property(
		__getfile,
		__setfile,
		__delfile,
		"""
		<doc:par>The filename without the path, i.e. the name part of the last component of
		<pyref property="path_segments"><property>path_segments</property></pyref>.
		The <lit>baz.html</lit> part of
		<lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getext(self):
		ext = None
		if len(self.path_segments):
			path_segment = self.path_segments[-1]
			name = path_segment[0]
			pos = name.rfind(".")
			if pos != -1:
				ext = name[pos+1:]
		return ext
	def __setext(self, ext):
		"""
		<doc:par>Setting the extension preserves the parameter in the last segment.</doc:par>
		"""
		if ext is None:
			del self.ext
		elif len(self.path_segments):
			path_segment = self.path_segments[-1]
			name = path_segment[0]
			pos = name.rfind(".")
			if pos != -1:
				name = name[:pos+1] + ext
			else:
				name = name + "." + ext
			if len(path_segment)>1:
				self.path_segments[-1] = (name, path_segment[1])
			else:
				self.path_segments[-1] = (name, )
	def __delext(self):
		"""
		<doc:par>Deleting the extension preserves the parameter in the last segment.</doc:par>
		"""
		if len(self.path_segments):
			path_segment = self.path_segments[-1]
			name = path_segment[0]
			pos = name.rfind(".")
			if pos != -1:
				name = name[:pos]
				if len(path_segment)>1:
					self.path_segments[-1] = (name, path_segment[1])
				else:
					self.path_segments[-1] = (name, )
	ext = property(
		__getext,
		__setext,
		__delext,
		"""
		<doc:par>The filename extension of the last segment of the path. The <lit>html</lit> part of
		<lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.</doc:par>
		"""
	)

	def __getquery(self):
		return self._query
	def __setquery(self, query):
		self._query = query
	def __delquery(self):
		self._query = None
	query = property(
		__getquery,
		__setquery,
		__delquery,
		"""
		<doc:par>The query component, i.e. the <lit>spam=eggs</lit> part of
		<lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.
		"""
	)

	def __getopaque_part(self):
		return self._opaque_part
	def __setopaque_part(self, opaque_part):
		self._opaque_part = opaque_part
	def __delopaque_part(self):
		self._opaque_part = None
	opaque_part = property(
		__getopaque_part,
		__setopaque_part,
		__delopaque_part,
		"""
		<doc:par>The opaque part (for schemes like <lit>mailto</lit> that are not
		hierarchical).</doc:par>
		"""
	)

	def __getfragment(self):
		return self._fragment
	def __setfragment(self, fragment):
		self._fragment = fragment
	def __delfragment(self):
		self._fragment = None
	fragment = property(
		__getfragment,
		__setfragment,
		__delfragment,
		"""
		<doc:par>The fragment identifier, which references a part of the resource,
		i.e. the <lit>frag</lit> part of
		<lit>http://user@www.foo.com/bar/baz.html;xyzzy?spam=eggs#frag</lit>.
		""")

	def __geturl(self):
		"""
		<doc:par>Getting <property>url</property> reassembles the &url;
		from the components.</doc:par>
		"""
		result = ""
		if self.scheme is not None:
			result += self.scheme + ":"
		if self.reg.usehierarchy:
			if self.authority is not None:
				result += "//" + self.authority
			result += self.path
			if self.query is not None:
				result += "?" + self.query
		else:
			result += self.opaque_part
		if self.reg.usefragment and self.fragment is not None:
			result += "#" + self.fragment
		return result
	def __seturl(self, url):
		"""
		<doc:par>Setting <property>url</property> parses <arg>url</arg>
		into the components. <arg>url</arg> may also be an <class>URL</class> instance,
		in which case the &url; will be copied.</doc:par>
		"""
		self.__clear()
		if url is None:
			return
		elif isinstance(url, URL):
			self.scheme = url.scheme
			self.userinfo = url.userinfo
			self.host = url.host
			self.port = url.port
			self.isabspath = url.isabspath
			self.path_segments = url.path_segments[:]
			self.reg_name = url.reg_name
			self.opaque_part = url.opaque_part
			self.query = url.query
			self.fragment = url.fragment
		else:
			if isinstance(url, unicode):
				url = _escape(url)
			# find the scheme (RFC2396, Section 3.1)
			pos = url.find(":")
			if pos != -1:
				scheme = url[:pos]
				if self._checkscheme(scheme): # if the scheme is illegal assume there is none (e.g. "/foo.php?x=http://www.bar.com", will *not* have the scheme "/foo.php?x=http")
					self.scheme = scheme # the info about what we have to expect in the rest of the URL can be found in self.reg now
					url = url[pos+1:]

			# find the fragment (RFC2396, Section 4.1)
			if self.reg.usefragment:
				# the fragment itself may not contain a "#", so find the last "#"
				pos = url.rfind("#")
				if pos != -1:
					self.fragment = _unescape(url[pos+1:])
					url = url[:pos]

			if self.reg.usehierarchy:
				# find the query (RFC2396, Section 3.4)
				pos = url.rfind("?")
				if pos != -1:
					self.query = _unescape(url[pos+1:])
					url = url[:pos]
				if url.startswith("//"):
					url = url[2:]
					# find the authority part (RFC2396, Section 3.2)
					pos = url.find("/")
					if pos!=-1:
						authority = url[:pos]
						url = url[pos:] # keep the "/"
					else:
						authority = url
						url = ""
					self.authority = authority
				self.path = url
			else:
				self.opaque_part = url
	def __delurl(self):
		"""
		<doc:par>After deleting the &url; the resulting object will refer
		to the <z>current document</z>.</doc:par>
		"""
		self.__clear()
	url = property(
		__geturl,
		__seturl,
		__delurl,
		"""
		<doc:par>The complete &url;</doc:par>
		"""
	)

	def withExt(self, ext):
		"""
		<doc:par>Return a new <class>URL</class> where the filename extension
		has been replaced with <arg>ext</arg>.</doc:par>
		"""
		url = URL(self)
		url.ext = ext
		return url

	def withoutExt(self):
		"""
		<doc:par>Return a new <class>URL</class> where the filename extension
		has been removed.</doc:par>
		"""
		url = URL(self)
		del url.ext
		return url

	def withFile(self, file):
		"""
		<doc:par>Return a new <class>URL</class> where the filename (i.e. the <rep>name</rep>
		of last component of 
		<pyref property="path_segments"><property>path_segments</property></pyref>)
		has been replaced with <arg>filename</arg>.</doc:par>
		"""
		url = URL(self)
		url.file = file
		return url

	def withFragment(self, fragment):
		"""
		<doc:par>Return a new <class>URL</class> where the fragment
		has been replaced with <arg>fragment</arg>.</doc:par>
		"""
		url = URL(self)
		url.fragment = fragment
		return url

	def __div__(self, otherurl):
		"""
		<doc:par>join <self/> with a another (possible relative) <class>URL</class> <arg>otherurl</arg>,
		to form a new <class>URL</class>.</doc:par>
		<doc:par><arg>otherurl</arg> may be a <class>str</class>, <class>unicode</class> or
		<class>URL</class> instance. It may be <lit>None</lit> (referring to the <z>current document</z>)
		in which case <self/> will be returned. It may also be a list. For this case a list
		will be returned where <method>__div__</method> will be applied to every item in the list. E.g.
		the following expression returns all the files in the current directory as absolute
		<class>URL</class>s (see <pyref method="files"><method>files</method></pyref> and
		<pyref function="here"><function>here</function></pyref> for further explanations):</doc:par>
		<doc:programlisting>
		url.here()/url.here().files()
		</doc:programlisting>
		"""
		if isinstance(otherurl, list):
			return [ self/url for url in otherurl ] # this makes url/list possible
		otherurl = URL(otherurl)
		newurl = URL()
		# RFC2396, Section 5.2 (2)
		if otherurl.scheme is None and otherurl.authority is None and otherurl.path=="" and otherurl.query is None:
			newurl = URL(self)
			newurl.fragment = otherurl.fragment
			return newurl
		if not self.reg.usehierarchy: # e.g. "mailto:x@y"/"file:foo"
			return otherurl
		# In violation of RFC2396 we treat file URLs as relative ones (if the base is a local URL)
		if otherurl.scheme=="file" and self.isLocal():
			del otherurl.scheme
			del otherurl.authority
		# RFC2396, Section 5.2 (3)
		if otherurl.scheme is not None:
			return otherurl
		newurl.scheme = self.scheme
		newurl.query = otherurl.query
		newurl.fragment = otherurl.fragment
		# RFC2396, Section 5.2 (4)
		if otherurl.authority is None:
			newurl.authority = self.authority
			# RFC2396, Section 5.2 (5)
			if otherurl.isabspath:
				newurl.path_segments = otherurl.path_segments[:]
				newurl.isabspath = 1
			else:
				# RFC2396, Section 5.2 (6) (a)
				path_segments = self.path_segments[:-1]
				# RFC2396, Section 5.2 (6) (b)
				path_segments.extend(otherurl.path_segments)
				# the following should be equivalent to RFC2396, Section 5.2 (6) (c)-(f)
				newurl.path_segments = _normalizePath(path_segments)
				newurl.isabspath = self.isabspath
		else:
			newurl.authority = otherurl.authority
			newurl.path_segments = otherurl.path_segments[:]
			newurl.isabspath = otherurl.isabspath
		return newurl

	def relative(self, baseurl):
		"""
		<doc:par>Return an relative <class>URL</class> <rep>rel</rep> such that
		<lit><arg>baseurl</arg>/<rep>rel</rep> == <self/></lit>, i.e. this is the
		inverse operation of <pyref method="__div__"><method>__div__</method></pyref>.</doc:par>
		<doc:par>If <self/> is relative, has a different
		<pyref property="scheme"><property>scheme</property></pyref> or
		<pyref property="authority"><property>authority</property></pyref> than <arg>baseurl</arg>
		or a non-hierarchical scheme, an identical copy of <self/> will be returned.</doc:par>
		"""
		# if self is relative don't do anything
		if self.scheme is None:
			return URL(self)
		# javascript etc.
		if not self.reg.usehierarchy:
			return URL(self)
		baseurl = URL(baseurl) # clone/coerce
		# only calculate a new URL if to the same server, else use the original
		if self.scheme != baseurl.scheme or self.authority != baseurl.authority:
			return URL(self)
		newurl = URL(self) # clone
		del newurl.scheme
		del newurl.authority
		selfpath_segments = _normalizePath(self.path_segments)
		basepath_segments = _normalizePath(baseurl.path_segments)
		while len(selfpath_segments)>1 and len(basepath_segments)>1 and selfpath_segments[0]==basepath_segments[0]:
			del selfpath_segments[0]
			del basepath_segments[0]
		# does the URL go to the same file?
		if selfpath_segments==basepath_segments and self.query==baseurl.query:
			# only return the fragment
			del newurl.path_segments
			del newurl.query
		else:
			# build a path from one file to the other
			selfpath_segments[:0] = [("..",)]*(len(basepath_segments)-1)
			if not len(selfpath_segments) or selfpath_segments==[("",)]:
				selfpath_segments = [(".",), ("",)]
			newurl.path_segments = selfpath_segments
		newurl.isabspath = 0
		return newurl

	def __str__(self):
		return self.url

	def __unicode__(self):
		return self.url

	def __repr__(self):
		return "URL(%r)" % self.url

	def __nonzero__(self):
		"""
		<doc:par>Return whether the <class>URL</class> is not empty, i.e. whether
		it is not the <class>URL</class> referring to the start of the current document.</doc:par>
		"""
		return self.url != ""

	def __eq__(self, other):
		"""
		<doc:par>Return whether two <pyref class="URL"><class>URL</class></pyref> instances are equal.
		Note that only properties relevant for the current scheme will be compared.</doc:par>
		"""
		if self.__class__!=other.__class__:
			return 0
		if self.scheme!=other.scheme:
			return 0
		if self.reg.usehierarchy:
			if self.reg.useserver:
				selfport = self.port or self.reg.defaultport
				otherport = other.port or other.reg.defaultport
				if self.userinfo!=other.userinfo or self.host!=other.host or selfport!=otherport:
					return 0
			else:
				if self.reg_name!=other.reg_name:
					return 0
			if self.isabspath!=other.isabspath or self.path_segments != other.path_segments:
				return 0
		else:
			if self.opaque_part!=other.opaque_part:
				return 0
		if self.query != other.query:
			return 0
		if self.fragment != other.fragment:
			return 0
		return 1

	def __ne__(self, other):
		"""
		<doc:par>Return whether two <pyref class="URL"><class>URL</class></pyref>s are <em>not</em>
		equal.</doc:par>
		"""
		return not self==other

	def __hash__(self):
		"""
		<doc:par>Return a hash value for <self/>, to be able to use <class>URL</class>s as
		dictionary keys. You must be careful not to modify an <class>URL</class> as soon
		as you use it as a dictionary key.</doc:par>
		"""
		res = hash(self.scheme)
		if self.reg.usehierarchy:
			if self.reg.useserver:
				res ^= hash(self.userinfo)
				res ^= hash(self.host)
				res ^= hash(self.port or self.reg.defaultport)
			else:
				res ^= hash(self.reg_name)
			res ^= hash(self.isabspath)
			res ^= hash(tuple(self.path_segments))
		else:
			res ^= hash(self.opaque_part)
		res ^= hash(self.query)
		res ^= hash(self.fragment)
		return res

	def isLocal(self):
		"""
		<doc:par>return whether <self/> refers to a local file, i.e. whether
		<self/> is a relative <class>URL</class> or the scheme is
		<lit>root</lit> or <lit>file</lit>).</doc:par>
		"""
		return self.reg.isLocal

	def asFilename(self):
		"""
		<doc:par>Return <self/> as a filename (which will only works if <self/>
		is local (see <pyref method="isLocal"><method>isLocal</method></pyref>)).</doc:par>
		"""
		if not self.isLocal():
			raise ValueError("cannot convert %r to filename: non local scheme %s not supported" % (self, self.scheme))
		path = "/".join([ segment[0] for segment in self.path_segments])
		if self.isabspath:
			path = "/" + path
		return urllib.url2pathname(self.path)

	def walk(self, dirsbefore=0, dirsafter=0, files=1):
		"""
		<doc:par>Recursively walk <self/> (if it is a local <class>URL</class>) and
		generate <class>URL</class>s for files and/or directories in <self/>. All <class>URL</class>s
		will be relative to <self/>.</doc:par>
		<doc:par>Arguments have the following meaning:</doc:par>
		<doc:ulist>
		<doc:item><arg>dirsbefore</arg>: Yield directory <class>URL</class>s before
		entering the directory?</doc:item>
		<doc:item><arg>dirsafter</arg>: Yield directory <class>URL</class>s after
		entering the directory?</doc:item>
		<doc:item><arg>files</arg>: Yield files <class>URL</class>s?</doc:item>
		</doc:ulist>
		"""
		for childurl in self.__walk(base=None, dirsbefore=dirsbefore, dirsafter=dirsafter, files=files):
			yield childurl

	def __walk(self, base=None, dirsbefore=0, dirsafter=0, files=1):
		"""
		<doc:par>Internal helper method for <pyref method="walk"><method>walk</method></pyref>.</doc:par>
		"""
		path = self.asFilename()
		if os.path.isdir(path):
			allfiles = os.listdir(path)
			allfiles.sort()
			url = Dirname(path) # adds a trailing / if neccessary
			if base is None:
				base = url
			if dirsbefore:
				yield url.relative(base)
			for file in allfiles:
				file = Filename(file) # the filename might include #, ? or ;
				for childurl in (url/URL(file)).__walk(base=base, dirsbefore=dirsbefore, dirsafter=dirsafter, files=files):
					yield childurl
			if dirsafter:
				yield url.relative(base)
		else:
			if base is None:
				base = self
			if files: # if we have to report files, do it
				yield self.relative(base)

	def files(self):
		"""
		<doc:par>Return a list of all files in the directory <self/> (which must be a
		local <class>URL</class>). If <self/> is a file, a list with <self/> as the only
		entry will be returned. As with <pyref method="walk"><method>walk</method></pyref>
		all <class>URL</class>s will be relative to <self/>.</doc:par>
		"""
		return [ url for url in self.walk(dirsbefore=0, dirsafter=0, files=1) ]

	def dirs(self):
		"""
		<doc:par>Return a list of all directories in the directory <self/> (which must be a
		local <class>URL</class>). If <self/> is a file, an empty list will be returned.
		As with <pyref method="walk"><method>walk</method></pyref>
		all <class>URL</class>s will be relative to <self/>.</doc:par>
		"""
		return [ url for url in self.walk(dirsbefore=1, dirsafter=0, files=0) ]

	def openread(self, headers=None, data=None):
		"""
		<doc:par>Open <self/> for reading. This returns a 
		<pyref class="ReadResource"><class>ReadResource</class></pyref> instance.
		See its <pyref class="ReadResource" method="__init__">constructor</pyref>
		for information about the arguments.</doc:par>
		"""
		return ReadResource(self, headers, data)

	def openwrite(self):
		"""
		<doc:par>Open <self/> for writing. This returns a 
		<pyref class="WriteResource"><class>WriteResource</class></pyref> instance.</doc:par>
		"""
		return WriteResource(self)

class Resource(object):
	def __repr__(self):
		return "<%s instance for %r at 0x%x>" % (self.__class__.__name__, self.url, id(self))

	def _getimagesize(self):
		if not hasattr(self, "_imagesize"):
			import Image
			img = Image.open(StringIO.StringIO(self.resdata))
			self._imagesize = img.size
		return self._imagesize
	imagesize = property(_getimagesize, None, None, "the size of the image as a (x, y) tuple")

class ReadResource(Resource):
	"""
	<doc:par>A <class>file</class> like object from which the content of an
	<pyref class="URL"><class>URL</class></pyref> can be read.</doc:par>
	<doc:par><class>ReadResource</class> tries to defer opening the connection
	as long as possible, i.e. creating a <class>ReadResource</class> instance
	does not open a connection, only accessing the relevant properties will.</doc:par>
	"""
	def __init__(self, url, headers, data):
		"""
		<doc:par>Create a new resource object for reading.</doc:par>
		<doc:par><arg>headers</arg> (which must be a mapping) will be used
		for additional headers to be passed in the request and data
		(which also is a mapping) will be used as the data for a
		<lit>POST</lit> request.</doc:par>
		"""
		self.url = URL(url)
		if url.isLocal():
			self.filename = os.path.abspath(url.asFilename()) # protect against calls to os.chdir
		if headers is not None:
			headers = headers.copy()
		self.reqheaders = headers
		if data is not None:
			data = data.copy()
		self.reqdata = data

	def __open(self):
		"""
		<doc:par>internal helper that opens the resource on demand.</doc:par>
		"""
		if not hasattr(self, "_stream"):
			if hasattr(self, "filename"):
				self._stream = open(self.filename, "rb")
			else:
				data = self.reqdata
				if data is not None:
					data = urllib.urlencode(self.reqdata)
				headers = self.reqheaders
				if headers is None:
					headers = {}
				req = urllib2.Request(url=self.url.url, data=data, headers=headers)
				self._stream = urllib2.urlopen(req)
				self._resheaders = self._stream.info()
				self._mimetype = (self._resheaders.maintype, self._resheaders.subtype)
				l = self._resheaders.get("Content-Length")
				if l:
					l = int(l)
				self._contentlength = l
				m = self._resheaders.get("Last-Modified")
				if m is not None:
					m = DateTime.DateTimeFrom(m)
				self._lastmodified = m
			self._buffer = StringIO.StringIO() # here the bytes read from the stream will be cached

	def close(self):
		"""
		<doc:par>Close the resource</doc:par>
		"""
		if hasattr(self, "_stream"):
			self._stream.close()
			del self._stream

	def _getlastmodified(self):
		if not hasattr(self, "_lastmodified"):
			if hasattr(self, "filename"): # local file
				self._lastmodified = DateTime.mktime(time.gmtime(self.stats.st_mtime))
			else:
				self.__open()
		return self._lastmodified
	lastmodified = property(
		_getlastmodified,
		None,
		None,
		"""
		<doc:par>The last modification date of the resource as a <class>DateTime</class> object.</doc:par>
		"""
	)

	def _getcontentlength(self):
		if not hasattr(self, "_contentlength"):
			if hasattr(self, "filename"): # local file
				self._contentlength = self.stats.st_size
			else:
				self.__open()
		return self._contentlength
	contentlength = property(
		_getcontentlength,
		None,
		None,
		"""
		<doc:par>The size of the resource in bytes</doc:par>
		"""
	)

	def _getmimetype(self):
		if not hasattr(self, "_mimetype"):
			if hasattr(self, "filename"): # local file
				mimetype = mimetypes.guess_type(self.filename)[0]
				if mimetype is None:
					self._mimetype = ("text", "plain")
				else:
					self._mimetype = tuple(mimetype.split("/"))
			else:
				self.__open()
		return self._mimetype
	mimetype = property(
		_getmimetype,
		None,
		None,
		"""
		<doc:par>The &mime; type of the resource as a (maintype, subtype) tuple</doc:par>
		"""
	)

	def _getresheaders(self):
		if not hasattr(self, "_resheaders"):
			if hasattr(self, "filename"): # local file
				self._resheaders = mimetools.Message(
					StringIO.StringIO(
						"Content-Type: %s/%s\nContent-Length: %d\nLast-modified: %s\n" %
						(self.mimetype[0], self.mimetype[1], self.contentlength, email.Utils.formatdate(self.lastmodified.gmticks()))
					)
				)
			else:
				self.__open()
		return self._resheaders
	resheaders = property(
		_getresheaders,
		None,
		None,
		"""
		<doc:par>The headers of the response as a <class>mimetools.Message</class> object</doc:par>
		"""
	)

	def _getstats(self):
		if not hasattr(self, "_stats"):
			if hasattr(self, "filename"): # local file
				self._stats = os.stat(self.filename)
			else:
				self._stats = None
		return self._stats
	stats = property(
		_getstats,
		None,
		None,
		"""
		<doc:par>The result of a <function>os.stat</function> call
		(or <lit>None</lit> for a non-local resource)</doc:par>
		"""
	)

	def read(self, size=-1):
		self.__open()
		data = self._stream.read(size)
		self._buffer.write(data)
		return data

	def _getresdata(self):
		self.__open()
		self.read() # read the rest
		data = self._buffer.getvalue()
		self.close()
		return data
	resdata = property(
		_getresdata,
		None,
		None,
		"""
		<doc:par>The content of the resource as a byte string</doc:par>
		"""
	)

class WriteResource(Resource):
	def __init__(self, url):
		"""
		<doc:par>Create a new resource object for writing
		(only local <pyref class="URL"><class>URL</class></pyref>s are supported).</doc:par>
		"""
		self.url = URL(url)
		if not url.isLocal():
			raise ValueError("cannot open %r for writing: non local scheme %s not supported" % (self.url, self.url.scheme))
		self.filename = os.path.abspath(url.asFilename()) # protect against calls to os.chdir

	def __del__(self):
		self.close()

	def __open(self):
		if not hasattr(self, "_stream"):
			try:
				self._stream = open(self.filename, "wb")
			except IOError, ex:
				if ex[0] != 2: # didn't work for some other reason than a non existing directory
					raise
				(splitpath, splitname) = os.path.split(self.filename)
				if splitpath == "": # we don't really have a parent
					raise # we don't have a directory to make so pass the error on
				else:
					os.makedirs(splitpath)
					self._stream = open(self.filename, "wb")
			self._buffer = StringIO.StringIO() # here the bytes written to the stream will be cached

	def close(self):
		"""
		<doc:par>Close the resource</doc:par>
		"""
		if self.contentlength==0: # if we've never written to the file, do it now, so the empty file will be created
			self.write("")
		if hasattr(self, "_stream"):
			self._stream.close()
			del self._stream

	def _getlastmodified(self):
		if not hasattr(self, "_lastmodified"):
			self._lastmodified = None
		return self._lastmodified
	lastmodified = property(
		_getlastmodified,
		None,
		None,
		"""
		<doc:par>The last modification date of the resource as a <class>DateTime</class> object.</doc:par>
		"""
	)

	def _getcontentlength(self):
		if hasattr(self, "_buffer"):
			return len(self._buffer.getvalue())
		else:
			return 0
	contentlength = property(
		_getcontentlength,
		None,
		None,
		"""
		<doc:par>The size of the resource in bytes</doc:par>
		"""
	)

	def _getmimetype(self):
		mimetype = mimetypes.guess_type(self.filename)[0]
		if mimetype is None:
			self._mimetype = ("text", "plain")
		else:
			self._mimetype = tuple(mimetype.split("/"))
		return self._mimetype
	mimetype = property(
		_getmimetype,
		None,
		None,
		"""
		<doc:par>The &mime; type of the resource as a (maintype, subtype) tuple</doc:par>
		"""
	)

	def _getresheaders(self):
		if not hasattr(self, "_resheaders"):
			self._resheaders = mimetools.Message(
				StringIO.StringIO(
					"Content-Type: %s/%s\nContent-Length: %d\nLast-modified: %s\n" %
					(self.mimetype[0], self.mimetype[1], self.contentlength, email.Utils.formatdate(self.lastmodified.gmticks()))
				)
			)
		return self._resheaders
	resheaders = property(
		_getresheaders,
		None,
		None,
		"""
		<doc:par>The headers of the response as a <class>mimetools.Message</class> object</doc:par>
		"""
	)

	def _getstats(self):
		return None
	stats = property(
		_getstats,
		None,
		None,
		"""
		<doc:par>always <lit>None</lit></doc:par>
		"""
	)

	def write(self, data):
		self.__open()
		self._stream.write(data)
		self._buffer.write(data)
		self._lastmodified = DateTime.utc()
		self._contentlength = self._buffer.tell()
		try:
			del self._resheaders
		except AttributeError:
			pass

	def tell(self):
		return len(self.resdata)

	def _getresdata(self):
		"""
		<doc:par>Getting this property returns the bytes that have been written to the resouce so far.</doc:par>
		"""
		if hasattr(self, "_buffer"):
			return self._buffer.getvalue()
		else:
			return ""
	def _setresdata(self, data):
		"""
		<doc:par>Setting this property will write <arg>data</arg> to the resource.</doc:par>
		"""
		if hasattr(self, "_stream") and self.contentlength:
			raise ValueError("resource %r has already been written to")
		self.__open()
		self.write(data)
		self.close()
	resdata = property(
		_getresdata,
		_setresdata,
		None,
		"""
		<doc:par>The content of the resource as a byte string</doc:par>
		"""
	)

warnings.filterwarnings("always", module="url")
